from typing import Optional, List

import click
from click.core import Context, Option

from .base import CONTEXT_SETTINGS, run_test, _RESULT_TYPES, _DEFAULT_RESULT_TYPE
from ...config.meta import __TITLE__, __VERSION__, __AUTHOR__, __AUTHOR_EMAIL__


# noinspection PyUnusedLocal
def print_version(ctx: Context, param: Option, value: bool) -> None:
    """
    Print version information of cli.

    :param ctx: click context
    :param param: current parameter's metadata
    :param value: value of current parameter
    """
    if not value or ctx.resilient_parsing:
        return
    click.echo('{title}, version {version}.'.format(title=__TITLE__.capitalize(), version=__VERSION__))
    click.echo('Developed by {author}, {email}.'.format(author=__AUTHOR__, email=__AUTHOR_EMAIL__))
    ctx.exit()


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--version', is_flag=True,
              callback=print_version, expose_value=False, is_eager=True,
              help="Show package's version information.")
@click.option('-i', '--input', 'input_content', type=str, help='Input content of special judge.')
@click.option('-o', '--output', 'output_content', type=str, help='Output content of special judge')
@click.option('-I', '--input_file', type=click.Path(exists=True, dir_okay=False, readable=True),
              help='Input file of special judge (if -i is given, this will be ignored).')
@click.option('-O', '--output_file', type=click.Path(exists=True, dir_okay=False, readable=True),
              help='Output file of special judge (if -o is given, this will be ignored).')
@click.option('-V', '--value', type=str, multiple=True,
              help='Attached values for special judge (do not named as "stdin" or "stdout").')
@click.option('-t', '--type', 'result_type',
              type=click.Choice(_RESULT_TYPES),
              help='Type of the final result.', default=_DEFAULT_RESULT_TYPE, show_default=True)
@click.option('-s', '--spj', type=str, required=True,
              help='Special judge script to be used.')
@click.option('-p', '--pretty', type=bool, is_flag=True,
              help='Use pretty mode to print json result.')
def cli(input_content: Optional[str], output_content: Optional[str],
        input_file: Optional[str], output_file: Optional[str],
        value: List[str], result_type: str, spj: str, pretty: bool):
    """
    Command line entry of pyspj CLI.

    :param input_content: Input content, should be a string or ``None``.
    :param output_content: Output content, should be a string or ``None``.
    :param input_file: Input filename, should be a string or ``None``.
    :param output_file: Output filename, should be a string or ``None``.
    :param value: Additional arguments, like ``KEY=VALUE``.
    :param result_type: Result type, should be a string.
    :param spj: Special judge string.
    :param pretty: Enable pretty print or not, default is on.
    """
    return run_test(input_content, output_content, input_file, output_file, value, result_type, spj, pretty)



import json

from .exception import _raise_exception_with_exit_code
from .values import _load_from_values
from ..script import execute_spj
from ..script.decorator import stdin_string_support, stdin_file_trans, stdout_string_support, stdout_file_trans
from ...models.general import ResultType

CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help']
)

_RESULT_TYPES = {item.lower() for item in ResultType.__members__.keys()}
_DEFAULT_RESULT_TYPE = ResultType.FREE.name.lower()


def run_test(input_content, output_content,
             input_file, output_file, value, result_type,
             spj, pretty):
    if not input_content and not input_file:
        _raise_exception_with_exit_code(1, 'Either -i or -I should be given.')
    if not output_content and not output_file:
        _raise_exception_with_exit_code(1, 'Either -o or -O should be given.')

    _execute_func = execute_spj
    if input_content:
        _execute_func = stdin_string_support(_execute_func)
        _input = input_content
    else:
        _execute_func = stdin_file_trans(_execute_func)
        _input = input_file

    if output_content:
        _execute_func = stdout_string_support(_execute_func)
        _output = output_content
    else:
        _execute_func = stdout_file_trans(_execute_func)
        _output = output_file

    result_type = ResultType.loads(result_type)
    result = _execute_func(
        spj=spj, type_=result_type,
        stdin=_input, stdout=_output,
        arguments=_load_from_values(list(value)),
    )

    print(json.dumps(result.to_json(), indent=4 if pretty else None, sort_keys=True))



import io
import sys
import traceback
from functools import wraps

from .decorator import string_support, file_trans
from .imports import _load_func
from ...models import load_result, SPJResult


def _spj_preprocess(spj):
    """
    Process spj to be safe.

    :param spj: Original special judge function.
    :return: Processed special judge function.
    """
    spj = _load_func(spj)

    @wraps(spj)
    def _func(*args, **kwargs):
        try:
            return spj(*args, **kwargs)
        except Exception as err:
            with io.StringIO() as eis:
                traceback.print_exception(*sys.exc_info(), file=eis)
                err_info = eis.getvalue()

            return False, 'Exception occurred while special judge - {cls}.'.format(cls=repr(err)), err_info

    return _func


def execute_spj(spj, stdin, stdout, type_=None, arguments=None) -> SPJResult:
    """
    Execute the special judge.

    :param spj: Special judge function, can be a string or a spj-function.
    :param stdin: Stdin stream, can be a string stream or a file stream.
    :param stdout: Stdout stream, should be a string stream or a file stream.
    :param type_: Load type, should be a result mode string or enum.
    :param arguments: Attached arguments.
    :return: Result of this special judge.
    """
    return load_result(_spj_preprocess(spj)(stdin=stdin, stdout=stdout, **(arguments or {})), type_)


def execute_spj_from_string(spj, stdin, stdout, type_=None, arguments=None) -> SPJResult:
    """
    Execute special judge with string value.

    :param spj: Special judge function, can be a string or a spj-function.
    :param stdin: Stdin stream, can be a string stream.
    :param stdout: Stdout stream, should be a string stream.
    :param type_: Load type, should be a result mode string or enum.
    :param arguments: Attached arguments.
    :return: Result of this special judge.
    """
    return execute_spj(string_support(_spj_preprocess(spj)), stdin, stdout, type_, arguments)


def execute_spj_from_file(spj, stdin_file, stdout_file, type_=None, arguments=None) -> SPJResult:
    """
    Execute special judge with string file.

    :param spj: Special judge function, can be a string or a spj-function.
    :param stdin_file: Stdin stream, can be a file stream.
    :param stdout_file: Stdout stream, should be a file stream.
    :param type_: Load type, should be a result mode string or enum.
    :param arguments: Attached arguments.
    :return: Result of this special judge.
    """
    return execute_spj(file_trans(_spj_preprocess(spj)), stdin_file, stdout_file, type_, arguments)



from typing import Optional, Mapping

from hbutils.model import get_repr_info
from hbutils.string import truncate

from .base import SPJResult


def _check_score(score: float) -> float:
    score = float(score)
    if score < 0:
        raise ValueError('Score value should be no less than 0 but {actual} found.'.format(actual=repr(score)))
    elif score > 1:
        raise ValueError('Score value should be no greater than 1 but {actual} found.'.format(actual=repr(score)))

    return score


class ContinuitySPJResult(SPJResult):
    """
    Overview:
        Result of continuity special judge.
    """

    def __init__(self, correctness: bool, score: float,
                 message: Optional[str] = None, detail: Optional[str] = None, **kwargs):
        """
        Constructor for :class:`pyspj.models.continuity.ContinuitySPJResult`.

        :param correctness: correctness of result
        :param score: score of result
        :param message: message of result
        :param detail: detail of result
        """
        SPJResult.__init__(self, correctness, message, detail, **kwargs)
        self.__score = _check_score(score)

    @property
    def score(self) -> float:
        """
        Score of this result.
        """
        return self.__score

    def to_json(self) -> Mapping[str, str]:
        return {
            **SPJResult.to_json(self),
            'score': self.score,
        }

    def __repr__(self):
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('correctness', lambda: repr(self.correctness)),
                ('score', lambda: '%.3f' % self.score),
                ('message', lambda: truncate(repr(self.message), width=64, tail_length=16, show_length=True)),
            ]
        )

    def _tuple(self):
        _correctness, _message, _detail = SPJResult._tuple(self)
        return _correctness, self.__score, _message, _detail



from typing import Mapping

from ...utils import list_to_envs

_FORBIDDEN_NAMES = {'stdin', 'stdout'}


def _check_values(values: Mapping[str, str]) -> Mapping[str, str]:
    """
    check value dict
    :param values: value dict
    :return: value dict
    """
    for name in _FORBIDDEN_NAMES:
        if name in values.keys():
            raise KeyError('Value should not be named as {name}.'.format(name=repr(name)))

    return values


def _load_from_values(values) -> Mapping[str, str]:
    """
    load values from list of strings
    :param values: list of strings
    :return: string dict
    """
    return _check_values(list_to_envs(values))



from hbutils.reflection import import_object

_DEFAULT_FUNC_NAME = '__spj__'


def _load_func_from_string(string: str):
    splits = string.split(":", maxsplit=1)
    if len(splits) == 1:
        package_name, func_name = splits[0], _DEFAULT_FUNC_NAME
    else:
        package_name, func_name = splits

    return import_object(func_name, package_name)


def _load_func(data):
    """
    load function from any kind of data
    :param data: raw data
    :return: loaded function
    """

    if isinstance(data, str):
        _func = _load_func_from_string(data)
    else:
        _func = data

    if not callable(_func):
        raise TypeError(
            'Callable type expected but {actual} found.'.format(actual=repr(type(_func).__name__)))

    return _func



from platform import python_version
from typing import Optional

import click
from click.core import Context, Option

from .base import CONTEXT_SETTINGS, _DEFAULT_RESULT_TYPE, run_test
from ...config.meta import __TITLE__, __VERSION__
from ...models import ResultType


def pyspj_entry(name: str, spj, result_type: str = _DEFAULT_RESULT_TYPE,
                version: Optional[str] = None, author: Optional[str] = None, email: Optional[str] = None):
    """
    Create your pyspj CLI entry.

    :param name: Name of the special judge.
    :param spj: Special judge function or string.
    :param result_type: Type of result, default is ``FREE``.
    :param version: Version information, default is ``None``.
    :param author: Author of this special judge, default is ``None``.
    :param email: Author email of this special judge, default is ``None``.
    :return: A click function, can be used directly to create a CLI program.
    """
    result_type = ResultType.loads(result_type)

    def print_version(ctx: Context, param: Option, value: bool) -> None:
        if not value or ctx.resilient_parsing:
            return
        if version:
            click.echo(f'Special judge - {name}, version {version}.')
        else:
            click.echo(f'Special judge - {name}.')
        if author:
            if email:
                click.echo(f'Developed by {author}, {email}.')
            else:
                click.echo(f'Developed by {author}.')
        click.echo(f'Powered by {__TITLE__}, version {__VERSION__}.')
        click.echo(f'Based on Python {python_version()}.')
        ctx.exit()

    @click.command(context_settings=CONTEXT_SETTINGS,
                   help=f"{name.capitalize()} - test a pair of given input and output.")
    @click.option('-v', '--version', is_flag=True, callback=print_version, expose_value=False, is_eager=True,
                  help="Show special judge's version information.")
    @click.option('-i', '--input', 'input_content', type=str, help='Input content of special judge.')
    @click.option('-o', '--output', 'output_content', type=str, help='Output content of special judge')
    @click.option('-I', '--input_file', type=click.Path(exists=True, dir_okay=False, readable=True),
                  help='Input file of special judge (if -i is given, this will be ignored).')
    @click.option('-O', '--output_file', type=click.Path(exists=True, dir_okay=False, readable=True),
                  help='Output file of special judge (if -o is given, this will be ignored).')
    @click.option('-V', '--value', type=str, multiple=True,
                  help='Attached values for special judge (do not named as "stdin" or "stdout").')
    @click.option('-p', '--pretty', type=bool, is_flag=True,
                  help='Use pretty mode to print json result.')
    def _built_cli(input_content, output_content,
                   input_file, output_file, value, pretty):
        return run_test(input_content, output_content, input_file, output_file, value, result_type, spj, pretty)

    return _built_cli



import warnings
from abc import ABCMeta
from typing import Optional, Mapping

from hbutils.model import get_repr_info
from hbutils.string import truncate

_DEFAULT_SUCCESS_MESSAGE = 'Congratulations, test passed!'
_DEFAULT_FAILURE_MESSAGE = 'Sorry, test failed!'


def _check_message(message: str) -> str:
    message = message.strip()
    _message_lines = message.splitlines()
    if len(_message_lines) > 1:
        warnings.warn('Message contains {n} lines, '
                      'we strongly recommend you to simplify the '
                      'message into exactly 1 line to keep its briefness.'.format(n=repr(len(_message_lines))))

    return message


class SPJResult(metaclass=ABCMeta):
    """
    Overview:
        Result of special judge.
        This is an abstract class.
    """

    def __init__(self, correctness: bool,
                 message: Optional[str] = None, detail: Optional[str] = None, **kwargs):
        """
        Constructor for :class:`pyspj.models.base.SPJResult`.

        :param correctness: correctness of result
        :param message: message of result
        :param detail: detail of result
        """
        self.__correctness = not not correctness
        self.__message = _check_message(
            message or (_DEFAULT_SUCCESS_MESSAGE if self.__correctness else _DEFAULT_FAILURE_MESSAGE))
        self.__detail = (detail or self.__message).rstrip()

    @property
    def correctness(self) -> bool:
        """
        Correctness of this result.
        """
        return self.__correctness

    @property
    def message(self) -> str:
        """
        Message of this result.
        """
        return self.__message

    @property
    def detail(self) -> str:
        """
        Detail information of this result.
        """
        return self.__detail

    def to_json(self) -> Mapping[str, str]:
        """
        Get JSON-formatted result data.

        :return: A json for the current result data.
        """
        return dict(
            correctness=self.correctness,
            message=self.message,
            detail=self.detail,
        )

    def __repr__(self):
        """
        Get representation format of current result object.

        :return: Representation string.
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('correctness', lambda: repr(self.correctness)),
                ('message', lambda: truncate(repr(self.message), width=64, tail_length=16, show_length=True))
            ]
        )

    def _tuple(self):
        return self.__correctness, self.__message, self.__detail

    def __eq__(self, other):
        if other is self:
            return True
        elif isinstance(other, self.__class__):
            return type(self) == type(other) and self._tuple() == other._tuple()
        else:
            return False

    def __hash__(self):
        return hash(self._tuple())



from typing import List, Mapping


def string_to_env(string: str):
    """
    Overview:
        Turn string ``KEY=VALUE`` to key-value tuple.

    Arguments:
        - string (:obj:`str`): Original key-value string.

    Returns:
        - (key, value): key and value

    Examples::
        >>> from pyspj.utils import string_to_env
        >>> string_to_env('THIS=1')
        ('THIS', '1')
        >>> string_to_env('THIS=1=2')
        ('THIS', '1=2')
    """
    _name, _value = string.split('=', maxsplit=1)
    return _name, _value


def list_to_envs(strings: List[str]) -> Mapping[str, str]:
    """
    Overview:
        Turn a list of strings to key-value dict.

    Arguments:
        - strings (:obj:`List[str]`): List of key-value strings.

    Returns:
        - envs (:obj:`Mapping[str, str]`): Separated key-value dict.

    Examples::
        >>> from pyspj.utils import list_to_envs
        >>> list_to_envs([
        ...     '1=2',
        ...     'THIS=',
        ...     'OK=kjsdlf-2398',
        ...     'THISX=1=2',
        ... ])
        {'1': '2', 'THIS': '', 'OK': 'kjsdlf-2398', 'THISX': '1=2'}
    """
    return {key: value for key, value in map(string_to_env, strings)}


def env_to_string(name, value) -> str:
    """
    Overview:
        Turn key and value to ``KEY=VALUE`` string.

    Arguments:
        - name (:obj:`str`): Name of key.
        - value (:obj:`str`): Value.

    Returns:
        - string (:obj:`str`): ``KEY=VALUE`` string.

    Examples::
        >>> from pyspj.utils import env_to_string
        >>> env_to_string('THIS', '')
        'THIS='
        >>> env_to_string('THIS', '1=2')
        'THIS=1=2'
    """
    return f'{name}={value}'


def envs_to_list(envs: Mapping[str, str]) -> List[str]:
    """
    Overview:
        Turn key-value dict to list of ``KEY=VALUE`` strings.

    Arguments:
        - envs (:obj:`Mapping[str, str]`): Key-value dict.

    Returns:
        - lenvs (:obj:`List[str]`): List of ``KEY=VALUE`` strings.

    Examples::
        >>> from pyspj.utils import envs_to_list
        >>> envs_to_list({
        ...     '1': '2',
        ...     'THIS': '',
        ...     'OK': 'kjsdlf-2398',
        ...     'THISX': '1=2',
        ... })
        ['1=2', 'THIS=', 'OK=kjsdlf-2398', 'THISX=1=2']
    """
    return [env_to_string(_name, _value) for _name, _value in envs.items()]



from .decorator import string_support, stdout_string_support, stdin_string_support, file_trans, stdout_file_trans, \
    stdin_file_trans
from .execute import execute_spj, execute_spj_from_file, execute_spj_from_string



import codecs
import io
from functools import wraps


def stdin_string_support(func):
    @wraps(func)
    def _func(stdin, **kwargs):
        if isinstance(stdin, str):
            with io.StringIO(stdin) as stdin_stream:
                return func(stdin=stdin_stream, **kwargs)
        else:
            return func(stdin=stdin, **kwargs)

    return _func


def stdout_string_support(func):
    @wraps(func)
    def _func(stdout, **kwargs):
        if isinstance(stdout, str):
            with io.StringIO(stdout) as stdout_stream:
                return func(stdout=stdout_stream, **kwargs)
        else:
            return func(stdout=stdout, **kwargs)

    return _func


def string_support(func):
    return stdin_string_support(stdout_string_support(func))


def stdin_file_trans(func):
    @wraps(func)
    def _func(stdin, **kwargs):
        with codecs.open(stdin, 'r') as stdin_file:
            return func(stdin=stdin_file, **kwargs)

    return _func


def stdout_file_trans(func):
    @wraps(func)
    def _func(stdout, **kwargs):
        with codecs.open(stdout, 'r') as stdout_file:
            return func(stdout=stdout_file, **kwargs)

    return _func


def file_trans(func):
    return stdin_file_trans(stdout_file_trans(func))



from enum import unique, IntEnum
from typing import Tuple, Optional

import enum_tools
from hbutils.model import int_enum_loads

from .base import SPJResult
from .continuity import ContinuitySPJResult
from .simple import SimpleSPJResult


def _load_from_values(correctness, score, message, detail) -> SPJResult:
    if score is None:
        return SimpleSPJResult(correctness, message, detail)
    else:
        return ContinuitySPJResult(correctness, score, message, detail)


def _load_result_from_dict(data: dict) -> SPJResult:
    _correctness = data['correctness']
    _score = data.get('score', None)
    _message = data.get('message', None)
    _detail = data.get('detail', None)

    return _load_from_values(_correctness, _score, _message, _detail)


def _load_core_result_from_tuple(head) -> Tuple[bool, Optional[float]]:
    if isinstance(head, (list, tuple)):
        _correctness, _score = tuple(head)
    else:
        _correctness, _score = head, None

    return _correctness, _score


def _load_result_from_tuple(data: tuple) -> SPJResult:
    if not data:
        raise ValueError(
            'Tuple result should has no less than 1 object but {actual} found.'.format(actual=repr(len(data))))
    elif len(data) == 1:
        (_core,), _message, _detail = data, None, None
    elif len(data) == 2:
        (_core, _message), _detail = data, None
    elif len(data) == 3:
        _core, _message, _detail = data
    else:
        raise ValueError(
            'Tuple result should has no more than 3 object but {actual} found.'.format(actual=repr(len(data))))

    _correctness, _score = _load_core_result_from_tuple(_core)
    return _load_from_values(_correctness, _score, _message, _detail)


def _load_result_from_any_data(data) -> SPJResult:
    if isinstance(data, SimpleSPJResult):
        return data
    elif isinstance(data, ContinuitySPJResult):
        return data
    elif isinstance(data, dict):
        return _load_result_from_dict(data)
    elif isinstance(data, (list, tuple)):
        return _load_result_from_tuple(tuple(data))
    else:
        return SimpleSPJResult(not not data)


@enum_tools.documentation.document_enum
@int_enum_loads(name_preprocess=str.upper)
@unique
class ResultType(IntEnum):
    """
    Overview:
        Loading types of the result.
    """
    FREE = 0  # doc: Free mode. Just use the returned result.
    SIMPLE = 1  # doc: Simple mode. Convert the returned result to simple spj result.
    CONTINUITY = 2  # doc: Continuity mode. Conbert the returned result to continuity spj result.


def load_result(data, type_=None) -> SPJResult:
    """
    Overview:
        Load the given result data with result type's format.

    Arguments:
        - data: The given original result data.
        - type\_: The expected type. Default is ``None``, which means ``ResultType.FREE``.

    Returns:
        - result (:obj:`SPJResult`): Converted result object.
    """
    _result = _load_result_from_any_data(data)
    type_ = ResultType.loads(type_ or ResultType.FREE)
    if type_ == ResultType.SIMPLE:
        return to_simple(_result)
    elif type_ == ResultType.CONTINUITY:
        return to_continuity(_result)
    else:
        return _result


def to_simple(data) -> SimpleSPJResult:
    """
    Overview:
        Transform the given data to :class:`pyspj.models.simple.SimpleSPJResult` object.

    Arguments:
        - data: The given original data.

    Returns:
        - result (:obj:`SimpleSPJResult`): Converted result object.
    """
    return SimpleSPJResult(**load_result(data).to_json())


def to_continuity(data) -> ContinuitySPJResult:
    """
    Overview:
        Transform the given data to :class:`pyspj.models.simple.ContinuitySPJResult` object.

    Arguments:
        - data: The given original data.

    Returns:
        - result (:obj:`ContinuitySPJResult`): Converted result object.
    """
    _dict = dict(load_result(data).to_json())
    if 'score' not in _dict:
        _dict['score'] = 0.0
    return ContinuitySPJResult(**_dict)
